# =============================================================================
# Minet Instagram API Scraper
# =============================================================================
#
# Instagram public API "scraper".
#
import json
from urllib.parse import quote
from ebbe import getpath
import re
from json.decoder import JSONDecodeError
from ural.instagram import (
    parse_instagram_url,
    is_instagram_username,
    is_instagram_post_shortcode,
    InstagramPost as ParsedInstagramPost,
    InstagramUser as ParsedInstagramUser,
    InstagramReel as ParsedInstagramReel,
)
from minet.constants import COOKIE_BROWSERS
from minet.utils import sleep_with_entropy
from minet.web import (
    create_pool_manager,
    create_request_retryer,
    grab_cookies,
    request,
    retrying_method,
    request,
)
from minet.instagram.constants import (
    INSTAGRAM_DEFAULT_THROTTLE,
    INSTAGRAM_MAX_RANDOM_ADDENDUM,
    INSTAGRAM_NB_REQUEST_LITTLE_WAIT,
    INSTAGRAM_DEFAULT_THROTTLE_LITTLE_WAIT,
    INSTAGRAM_MAX_RANDOM_ADDENDUM_LITTLE_WAIT,
    INSTAGRAM_NB_REQUEST_BIG_WAIT,
    INSTAGRAM_DEFAULT_THROTTLE_BIG_WAIT,
    INSTAGRAM_MAX_RANDOM_ADDENDUM_BIG_WAIT,
    INSTAGRAM_MIN_TIME_RETRYER,
    INSTAGRAM_URL,
    INSTAGRAM_PUBLIC_API_DEFAULT_TIMEOUT,
)
from minet.instagram.exceptions import (
    InstagramError500,
    InstagramPublicAPIInvalidResponseError,
    InstagramInvalidCookieError,
    InstagramInvalidTargetError,
    InstagramTooManyRequestsError,
    InstagramNoPublicationError,
    InstagramPrivateOrNonExistentAccountError,
    InstagramHashtagNeverUsedError,
    InstagramAccountNoFollowError,
    InstagramPrivateAccountError,
)
from minet.instagram.formatters import (
    format_comment,
    format_hashtag_post,
    format_post,
    format_user,
    format_user_info,
)

INSTAGRAM_GRAPHQL_ENDPOINT = "https://www.instagram.com/graphql/query/"
INSTAGRAM_HASHTAG_QUERY_HASH = "9b498c08113f1e09617a1703c22b2f32"
INSTAGRAM_USER_ID_QUERY_HASH = "c9100bf9110dd6361671f113dd02e7d6"
INSTAGRAM_MAGIC_TOKEN_PATTERN = re.compile(r"\"X-IG-App-ID\"\s*:\s*\"(\d+)\"")
INSTAGRAM_ID_PATTERN = re.compile(r"\d+")
INSTAGRAM_GET_USER_ID_PATTERN = re.compile(r"\d+_(\d+)")


def forge_comments_url(
    post,
    comment_id=None,
    min_or_max_id=None,
):
    if comment_id is not None and min_or_max_id is not None:
        url = "https://www.instagram.com/api/v1/media/%s/comments/%s/child_comments/?max_id=%s" % (
            post,
            comment_id,
            min_or_max_id
        )
        return url

    if min_or_max_id is not None:
        url = "https://www.instagram.com/api/v1/media/%s/comments/?min_id=%s&can_support_threading=true&permalink_enabled=false" % (
            post,
            min_or_max_id
        )
        return url

    url = "https://www.instagram.com/api/v1/media/%s/comments/?can_support_threading=true&permalink_enabled=false" % post
    return url


def forge_hashtag_search_url(name, cursor=None, count=50):
    params = {"tag_name": name, "first": count}

    if cursor is not None:
        params["after"] = cursor

    url = INSTAGRAM_GRAPHQL_ENDPOINT + "?query_hash=%s&variables=%s" % (
        INSTAGRAM_HASHTAG_QUERY_HASH,
        quote(json.dumps(params)),
    )

    return url


def forge_post_url_from_id(post_id):
    return "https://www.instagram.com/api/v1/media/%s/info/" % post_id


def forge_post_url_from_shortcode(post_shortcode):
    return "https://www.instagram.com/p/%s/?__a=1&__d=dis" % post_shortcode


def forge_username_url(user_id):
    params = {"user_id": user_id, "include_reel": True}

    url = INSTAGRAM_GRAPHQL_ENDPOINT + "?query_hash=%s&variables=%s" % (
        INSTAGRAM_USER_ID_QUERY_HASH,
        quote(json.dumps(params)),
    )

    return url


def forge_user_url(name):
    return "https://i.instagram.com/api/v1/users/web_profile_info/?username=%s" % name


def forge_user_followers_url(
    id,
    count=50,
    max_id=None,
):
    url = (
        "https://i.instagram.com/api/v1/friendships/%s/followers/?count=%s&search_surface=follow_list_page"
        % (
            id,
            count,
        )
    )
    if max_id:
        url += "&max_id=%s" % max_id

    return url


def forge_user_following_url(
    id,
    count=50,
    max_id=None,
):
    url = (
        "https://i.instagram.com/api/v1/friendships/%s/following/?count=%s&search_surface=follow_list_page"
        % (
            id,
            count,
        )
    )
    if max_id:
        url += "&max_id=%s" % max_id

    return url


def forge_user_posts_url(
    name,
    count=50,
    max_id=None,
):
    if max_id is not None:
        u = INSTAGRAM_GET_USER_ID_PATTERN.search(max_id)

        if u is not None:
            user = u.group(1)
            url = "https://i.instagram.com/api/v1/feed/user/%s/?count=%s&max_id=%s" % (
                user,
                count,
                max_id,
            )
            return url

    url = "https://i.instagram.com/api/v1/feed/user/%s/username/?count=%s" % (
        name,
        count,
    )
    return url


class InstagramAPIScraper(object):
    def __init__(self, cookie="firefox"):
        self.pool_manager = create_pool_manager(
            timeout=INSTAGRAM_PUBLIC_API_DEFAULT_TIMEOUT
        )

        if cookie in COOKIE_BROWSERS:
            get_cookie_for_url = grab_cookies(cookie)
            cookie = get_cookie_for_url(INSTAGRAM_URL)

        if not cookie:
            raise InstagramInvalidCookieError

        self.cookie = cookie
        self.magic_token = None
        self.nb_calls = 0
        self.retryer = create_request_retryer(
            min=INSTAGRAM_MIN_TIME_RETRYER,
            additional_exceptions=[InstagramTooManyRequestsError, InstagramError500],
        )

    @retrying_method()
    def request_json(self, url, magic_token=False):
        headers = {"Cookie": self.cookie}
        if magic_token:
            headers["X-IG-App-ID"] = self.magic_token

        response = request(
            url,
            pool_manager=self.pool_manager,
            spoof_ua=True,
            headers=headers,
            known_encoding="utf-8",
        )

        text = response.text()

        if (
            "the link you followed may be broken, or the page may have been removed"
            in text.lower().strip()
        ):
            raise InstagramInvalidTargetError

        try:
            data = json.loads(text)
        except JSONDecodeError:
            raise RuntimeError("HTML for the request to " + url + " : " + text)

        if response.status == 429:
            raise InstagramTooManyRequestsError

        if response.status == 500:
            raise InstagramError500

        if response.status >= 400:
            if data["message"].lower().strip() == "media not found or unavailable":
                raise InstagramInvalidTargetError

            raise InstagramPublicAPIInvalidResponseError(url, response.status, data)

        if self.nb_calls != 0 and (self.nb_calls % INSTAGRAM_NB_REQUEST_BIG_WAIT) == 0:
            sleep_with_entropy(
                INSTAGRAM_DEFAULT_THROTTLE_BIG_WAIT,
                INSTAGRAM_MAX_RANDOM_ADDENDUM_BIG_WAIT,
            )
            self.nb_calls += 1

            return data

        elif (
            self.nb_calls != 0
            and (self.nb_calls % INSTAGRAM_NB_REQUEST_LITTLE_WAIT) == 0
        ):

            sleep_with_entropy(
                INSTAGRAM_DEFAULT_THROTTLE_LITTLE_WAIT,
                INSTAGRAM_MAX_RANDOM_ADDENDUM_LITTLE_WAIT,
            )
            self.nb_calls += 1

            return data

        sleep_with_entropy(INSTAGRAM_DEFAULT_THROTTLE, INSTAGRAM_MAX_RANDOM_ADDENDUM)
        self.nb_calls += 1

        return data

    def get_magic_token(self):
        response = request("https://www.instagram.com/disney", known_encoding="utf-8")

        if response.status >= 400:
            return None

        t = INSTAGRAM_MAGIC_TOKEN_PATTERN.search(response.text())
        if t is None:
            return None

        self.magic_token = t.group(1)
        return self.magic_token

    def ensure_magic_token(method):
        def wrapped(self, *args, **kwargs):

            if self.magic_token is None:
                self.get_magic_token()

            return method(self, *args, **kwargs)

        return wrapped

    @ensure_magic_token
    def comments(self, post):
        if not INSTAGRAM_ID_PATTERN.match(post):
            parsed = parse_instagram_url(post)
            if isinstance(parsed, (ParsedInstagramPost, ParsedInstagramReel)):
                shortcode = parsed.id

            elif is_instagram_post_shortcode(post):
                shortcode = post

            else:
                raise InstagramInvalidTargetError

            url = forge_post_url_from_shortcode(shortcode)

            data_post = self.request_json(url, magic_token=True)
            if not data_post:
                raise InstagramInvalidTargetError

            post = getpath(data_post, ["items", 0, "pk"])

        min_id = None

        while True:
            url = forge_comments_url(post, min_or_max_id=min_id)

            data = self.request_json(url, magic_token=True)

            if not data:
                break

            items = data.get("comments")

            if not items:
                break

            for item in items:

                if item.get("type") == 2:
                    continue

                yield format_comment(item)

                if item.get("child_comment_count") > 0:

                    max_id = ""

                    while True:
                        url = forge_comments_url(post, comment_id=item.get("pk"), min_or_max_id=max_id)

                        data_comment = self.request_json(url, magic_token=True)

                        if not data_comment:
                            break

                        children_items = data_comment.get("child_comments")

                        for children_item in children_items:
                            yield format_comment(children_item)

                        more_available = data_comment.get("has_more_tail_child_comments")

                        if not more_available:
                            break

                        max_id = data_comment.get("next_max_child_cursor")

            min_id = data.get("next_min_id")

            if not min_id:
                break

    def search_hashtag(self, hashtag):
        hashtag = hashtag.lstrip("#")
        cursor = None

        while True:
            url = forge_hashtag_search_url(hashtag, cursor=cursor)

            data = self.request_json(url)

            if not getpath(data, ["data", "hashtag"]):
                raise InstagramHashtagNeverUsedError

            data = getpath(data, ["data", "hashtag", "edge_hashtag_to_media"])

            if not data:
                break

            edges = data.get("edges")

            for edge in edges:
                yield format_hashtag_post(edge["node"])

            has_next_page = getpath(data, ["page_info", "has_next_page"])

            if not has_next_page:
                break

            cursor = getpath(data, ["page_info", "end_cursor"])

    @ensure_magic_token
    def post_infos(self, name):
        if INSTAGRAM_ID_PATTERN.match(name):
            url = forge_post_url_from_id(name)

        else:
            parsed = parse_instagram_url(name)
            if isinstance(parsed, (ParsedInstagramPost, ParsedInstagramReel)):
                shortcode = parsed.id

            elif is_instagram_post_shortcode(name):
                shortcode = name

            else:
                raise InstagramInvalidTargetError

            url = forge_post_url_from_shortcode(shortcode)

        data = self.request_json(url, magic_token=True)

        if not data:
            raise InstagramInvalidTargetError

        return format_post(getpath(data, ["items", 0]))

    def get_username(self, name):
        if INSTAGRAM_ID_PATTERN.match(name):
            url = forge_username_url(name)

            data_user_id = self.request_json(url)

            if not data_user_id:
                raise InstagramInvalidTargetError

            name = getpath(data_user_id, ["data", "user", "reel", "user", "username"])

        else:
            parsed = parse_instagram_url(name)

            if isinstance(parsed, (ParsedInstagramPost, ParsedInstagramUser)):
                if not parsed.name:
                    raise InstagramInvalidTargetError

                name = parsed.name

            else:
                name = name.lstrip("@")

                if not is_instagram_username(name):
                    raise InstagramInvalidTargetError

        return name

    @ensure_magic_token
    def get_user(self, name):
        url = forge_user_url(name)

        return self.request_json(url, magic_token=True)

    @ensure_magic_token
    def user_followers(self, name):
        name = self.get_username(name)

        max_id = None
        data_user = self.get_user(name)
        user_id = getpath(data_user, ["data", "user", "id"])

        while True:
            url = forge_user_followers_url(user_id, max_id=max_id)

            data = self.request_json(url, magic_token=True)

            if not data:
                break

            items = data.get("users")

            if not items:
                followed_count = getpath(
                    data_user, ["data", "user", "edge_followed_by", "count"]
                )
                if followed_count:
                    raise InstagramPrivateAccountError(followed_count)

                raise InstagramAccountNoFollowError

            for item in items:
                yield format_user(item)

            max_id = data.get("next_max_id")

            if not max_id:
                break

    @ensure_magic_token
    def user_following(self, name):
        name = self.get_username(name)

        max_id = None
        data_user = self.get_user(name)
        user_id = getpath(data_user, ["data", "user", "id"])

        while True:
            url = forge_user_following_url(user_id, max_id=max_id)

            data = self.request_json(url, magic_token=True)

            if not data:
                break

            items = data.get("users")

            if not items:
                followers_count = getpath(
                    data_user, ["data", "user", "edge_follow", "count"]
                )
                if followers_count:
                    raise InstagramPrivateAccountError(followers_count)

                raise InstagramAccountNoFollowError

            for item in items:
                yield format_user(item)

            max_id = data.get("next_max_id")

            if not max_id:
                break

    @ensure_magic_token
    def user_posts(self, name):
        name = self.get_username(name)

        max_id = None

        while True:
            url = forge_user_posts_url(name, max_id=max_id)

            data = self.request_json(url, magic_token=True)

            if not data:
                break

            items = data.get("items")

            if not items:

                if getpath(data, ["user", "username"]):
                    raise InstagramNoPublicationError

                raise InstagramPrivateOrNonExistentAccountError

            for item in items:
                yield format_post(item)

            more_available = data.get("more_available")

            if not more_available:
                break

            max_id = data.get("next_max_id")

    @ensure_magic_token
    def user_infos(self, name):
        name = self.get_username(name)

        data = self.get_user(name)

        if not data:
            raise InstagramInvalidTargetError

        user = getpath(data, ["data", "user"])

        if not user:
            raise InstagramInvalidTargetError

        return format_user_info(user)



from minet.crawl.types import CrawlResult, CrawlJob, UrlOrCrawlJob
from minet.crawl.state import CrawlerState
from minet.crawl.spiders import (
    Spider,
    SpiderResult,
    SpiderNextJobs,
    FunctionSpider,
    DefinitionSpider,
    DefinitionSpiderOutput,
)
from minet.crawl.crawler import Crawler



import sys
from typing import Union
from os import PathLike
from io import FileIO
from bs4 import BeautifulSoup

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

AnyPath = Union[str, PathLike]
AnyFileTarget = Union[AnyPath, FileIO]
AnyScrapableTarget = Union[str, BeautifulSoup]



# =============================================================================
# Minet Extract Content CLI Action
# =============================================================================
#
# Logic of the extract action.
#
from typing import Optional, List, Dict, Iterator, Any

from os.path import isdir
from trafilatura.core import bare_extraction
from dataclasses import dataclass
from itertools import count
from casanova import TabularRecord, ThreadSafeEnricher

from minet.multiprocessing import LazyPool
from minet.fs import read_potentially_gzipped_path
from minet.cli.utils import (
    create_fetch_like_report_iterator,
    with_enricher_and_loading_bar,
    FetchReportLikeItem,
)
from minet.cli.console import console
from minet.cli.reporters import report_error
from minet.exceptions import TrafilaturaError
from minet.cli.exceptions import FatalError


@dataclass
class ExtractWorkerPayload:
    id: int
    encoding: Optional[str]
    path: Optional[str]
    text: Optional[str]


@dataclass
class ExtractResult:
    id: int
    error: Optional[Exception] = None
    meta: Optional[Dict[str, Any]] = None


@dataclass
class ExtractAddendum(TabularRecord):
    extract_error: Optional[str] = None
    canonical_url: Optional[str] = None
    title: Optional[str] = None
    description: Optional[str] = None
    content: Optional[str] = None
    comments: Optional[str] = None
    author: Optional[str] = None
    categories: Optional[List[str]] = None
    tags: Optional[List[str]] = None
    date: Optional[str] = None
    sitename: Optional[str] = None


def plural(meta: Dict[str, Any], key: str) -> List[str]:
    l = meta.get(key, []) or []

    items = []

    if not l:
        return items

    for item in l:
        if isinstance(item, dict):
            item = item.get("name")

            if item is None:
                continue

        if isinstance(item, (int, float)):
            item = str(item)

        if not isinstance(item, str):
            continue

        for subitem in item.split(","):
            subitem = subitem.strip()

            if subitem:
                items.append(subitem)

    return items


def worker(payload: ExtractWorkerPayload) -> ExtractResult:
    text = payload.text

    result = ExtractResult(id=payload.id)

    # Reading file
    if text is None:
        try:
            text = read_potentially_gzipped_path(
                payload.path, encoding=payload.encoding
            )
        except (FileNotFoundError, UnicodeDecodeError) as e:
            result.error = e
            return result

    # Attempting extraction
    try:
        # https://trafilatura.readthedocs.io/en/latest/corefunctions.html
        # TODO: discuss deduplication
        # TODO: fallback options
        result.meta = bare_extraction(text)
    except Exception as e:
        result.error = TrafilaturaError(reason=e)
        return result

    return result


@with_enricher_and_loading_bar(
    headers=ExtractAddendum,
    enricher_type="threadsafe",
    index_column="extract_original_index",
    title="Extracting text",
    unit="docs",
)
def action(cli_args, enricher: ThreadSafeEnricher, loading_bar):
    if cli_args.input_dir is not None and not isdir(cli_args.input_dir):
        raise FatalError(
            'Could not find the [cyan]-I/--input-dir "{}"[/cyan] directory!'.format(
                cli_args.input_dir
            )
        )

    items = create_fetch_like_report_iterator(cli_args, enricher)

    worked_on: Dict[int, FetchReportLikeItem] = {}

    def payloads() -> Iterator[ExtractWorkerPayload]:
        current_id = count()

        for item in items:

            # Items we cannot process
            if item.error is not None:
                loading_bar.advance()
                loading_bar.inc_stat(item.error, style="error")
                enricher.writerow(item.index, item.row)
                continue

            item_id = next(current_id)

            worked_on[item_id] = item

            yield ExtractWorkerPayload(
                id=item_id, encoding=item.encoding, path=item.path, text=item.text
            )

    pool = LazyPool(cli_args.processes)

    loading_bar.set_title("Extracting text (p=%i)" % pool.processes)

    warned_about_input_dir = False

    with pool:
        for result in pool.imap_unordered(
            worker, payloads(), chunksize=cli_args.chunk_size
        ):
            with loading_bar.step():
                assert isinstance(result, ExtractResult)

                item = worked_on.pop(result.id)
                row = item.row

                if result.error is not None:
                    error_code = report_error(result.error)
                    loading_bar.inc_stat(error_code, style="error")

                    if not warned_about_input_dir and error_code == "file-not-found":
                        warned_about_input_dir = True

                        if cli_args.input_dir is None:
                            console.warning(
                                "did you forget to give [cyan]-I/--input-dir[/cyan]? "
                            )
                        else:
                            console.warning(
                                'are you sure [cyan]-I/--input-dir "{}"[/cyan] is the correct path?'.format(
                                    cli_args.input_dir
                                )
                            )

                    addendum = ExtractAddendum(extract_error=error_code)
                    enricher.writerow(item.index, row, addendum)

                    continue

                meta = result.meta

                if meta is None:
                    loading_bar.inc_stat("no-trafilatura-result", style="error")
                    console.warning("did you forget to use [cyan]-i/--input[/cyan]?")
                    addendum = ExtractAddendum(extract_error="no-trafilatura-result")
                else:
                    addendum = ExtractAddendum(
                        canonical_url=meta.get("url"),
                        title=meta.get("title"),
                        description=meta.get("description"),
                        content=meta.get("text"),
                        comments=meta.get("comments"),
                        author=meta.get("author"),
                        categories=plural(meta, "categories"),
                        tags=plural(meta, "tags"),
                        date=meta.get("date"),
                        sitename=meta.get("sitename"),
                    )

                enricher.writerow(item.index, row, addendum)



from argparse import Action

from minet.cli.argparse import command, BooleanAction


class UrlFragmentAction(Action):
    """
    Custom argparse action to handle --dont-* flags and *except-routing flag.
    """

    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        super().__init__(option_strings, dest, nargs=0, **kwargs)

    def __call__(self, parser, cli_args, values, option_string=None):
        attr = (
            False
            if option_string.startswith("--no-") or option_string.startswith("--dont-")
            else True
        )

        if attr and option_string.endswith("except-routing"):
            attr = "except-routing"

        setattr(cli_args, self.dest, attr)


URL_PARSE_COMMAND = command(
    "url-parse",
    "minet.cli.url_parse.url_parse",
    title="Minet Url Parse Command",
    description="""
        Parse the urls contained in a CSV file using the python `ural`
        library to extract useful information about them such as their
        normalized version, domain name, etc.
    """,
    epilog="""
        Columns being added to the output:

        . "normalized_url": urls aggressively normalized by removing any part
          that is not useful to determine which resource it is actually
          pointing at.
        . "inferred_redirection": redirection directly inferred from the
          url without needing to make any HTTP request.
        . "domain_name": TLD-aware domain name of the url.
        . "hostname": full hostname of the url.
        . "normalized_hostname": normalized hostname, i.e. stripped of "www",
          "m" or some language subdomains etc., of the url.
        . "shortened": whether the url is probably shortened or
          not (bit.ly, t.co etc.).
        . "typo": whether the url probably contains typo or not
          (such as inclusive language in french : curieux.se etc.).
        . "homepage": whether the given url looks like a website's
          homepage.
        . "should_resolve": whether the given url looks like something
          we should resolve, i.e. shortened url.

        columns being added with --facebook:

        . "facebook_type": the type of Facebook resource symbolized by the
          parsed url (post, video etc.).
        . "facebook_id": Facebook resource id.
        . "facebook_full_id": Facebook full resource id.
        . "facebook_handle": Facebook handle for people, pages etc.
        . "facebook_normalized_url": normalized Facebook url.

        columns being added with --youtube:

        . "youtube_type": YouTube resource type (video, channel etc.).
        . "youtube_id": YouTube resource id.
        . "youtube_name": YouTube resource name.

        columns being added with --twitter:

        . "twitter_type": Twitter resource type (user or tweet).
        . "twitter_user_screen_name": Twitter user's screen name.
        . "tweet_id": id of tweet.

        Examples:

        . Creating a report about a file's urls:
            $ minet url-parse url -i posts.csv > report.csv

        . Keeping only selected columns from the input file:
            $ minet url-parse url -i posts.csv -s id,url,title > report.csv

        . Parsing Facebook urls:
            $ minet url-parse url -i fbposts.csv --facebook > report.csv

        . Parsing YouTube urls:
            $ minet url-parse url -i ytvideos.csv --youtube > report.csv

        . Parsing Twitter urls:
            $ minet url-parse url -i tweets.csv --twitter > report.csv
    """,
    variadic_input={"dummy_column": "url", "item_label": "url"},
    arguments=[
        {
            "flag": "--facebook",
            "help": "Whether to consider and parse the given urls as coming from Facebook.",
            "action": "store_true",
        },
        {
            "flag": "--twitter",
            "help": "Whether to consider and parse the given urls as coming from Twitter.",
            "action": "store_true",
        },
        {
            "flag": "--youtube",
            "help": "Whether to consider and parse the given urls as coming from YouTube.",
            "action": "store_true",
        },
        {
            "flags": [
                "--infer-redirection",
                "--dont-infer-redirection",
            ],
            "help": "Whether or not to attempt resolving common redirects by leveraging well-known GET parameters when normalizing url. Defaults to infer redirection.",
            "dest": "infer_redirection",
            "action": BooleanAction,
            "default": True,
        },
        {
            "flags": [
                "--fix-common-mistakes",
                "--dont-fix-common-mistakes",
            ],
            "help": "Whether or not to attempt to fix common URL mistakes when normalizing url. Defaults to fix common mistakes.",
            "dest": "fix_common_mistakes",
            "action": BooleanAction,
            "default": True,
        },
        {
            "flags": [
                "--normalize-amp",
                "--dont-normalize-amp",
            ],
            "help": "Whether or not to attempt to normalize Google AMP urls when normalizing url. Defaults to normalize amp.",
            "dest": "normalize_amp",
            "action": BooleanAction,
            "default": True,
        },
        {
            "flags": [
                "--quoted",
                "--no-quoted",
            ],
            "help": "Whether or not to normalize to a quoted or unquoted version of the url when normalizing url. Defaults to quoted.",
            "dest": "quoted",
            "action": BooleanAction,
            "default": True,
        },
        {
            "flags": [
                "--sort-query",
                "--dont-sort-query",
            ],
            "help": "Whether or not to sort query items when normalizing url. Defaults to sort query.",
            "dest": "sort_query",
            "action": BooleanAction,
            "default": True,
        },
        {
            "flags": [
                "--strip-authentication",
                "--dont-strip-authentication",
            ],
            "help": "Whether or not to strip authentication when normalizing url. Defaults to strip authentication.",
            "dest": "strip_authentication",
            "action": BooleanAction,
            "default": True,
        },
        {
            "flags": [
                "--strip-fragment",
                "--dont-strip-fragment",
                "--strip-fragment-except-routing",
            ],
            "help": "Whether or not to strip the url's fragment when normalizing url. If set to `--strip-fragment-except-routing`, will only strip the fragment if the fragment is not deemed to be js routing (i.e. if it contains a `/`). Defaults to strip fragment except routing.",
            "dest": "strip_fragment",
            "action": UrlFragmentAction,
            "default": "except-routing",
        },
        {
            "flags": [
                "--strip-index",
                "--dont-strip-index",
            ],
            "help": "Whether or not to strip trailing index when normalizing url. Defaults to strip index.",
            "dest": "strip_index",
            "action": BooleanAction,
            "default": True,
        },
        {
            "flags": [
                "--strip-irrelevant-subdomains",
                "--dont-strip-irrelevant-subdomains",
            ],
            "help": "Whether or not to strip trailing irrelevant-subdomains such as `www` etc. when normalizing url. Defaults to no strip irrelevantsubdomains.",
            "dest": "strip_irrelevant_subdomains",
            "action": BooleanAction,
            "default": True,
        },
        {
            "flags": [
                "--strip-lang-query-items",
                "--dont-strip-lang-query-items",
            ],
            "help": "Whether or not to strip language query items (ex: `gl=pt_BR`) when normalizing url. Defaults to no strip lang query items.",
            "dest": "strip_lang_query_items",
            "action": BooleanAction,
            "default": False,
        },
        {
            "flags": [
                "--strip-lang-subdomains",
                "--dont-strip-lang-subdomains",
            ],
            "help": "Whether or not to strip language subdomains (ex: `fr-FR.lemonde.fr` to only `lemonde.fr` because `fr-FR` isn't a relevant subdomain, it indicates the language and the country) when normalizing url. Defaults to no strip lang subdomains.",
            "dest": "strip_lang_subdomains",
            "action": BooleanAction,
            "default": False,
        },
        {
            "flags": ["--strip-protocol", "--dont-strip-protocol"],
            "help": "Whether or not to strip the protocol when normalizing the url. Defaults to strip protocol.",
            "dest": "strip_protocol",
            "action": BooleanAction,
            "default": True,
        },
        {
            "flags": [
                "--strip-trailing-slash",
                "--dont-strip-trailing-slash",
            ],
            "help": "Whether or not to trailing slash when normalizing url. Defaults to strip trailing slash.",
            "dest": "strip_trailing_slash",
            "action": BooleanAction,
            "default": True,
        },
    ],
)



# =============================================================================
# Minet CrowdTangle Posts By Id CLI Action
# =============================================================================
#
# Logic of the `ct posts-by-id` action.
#
from ural import ensure_protocol
from ural.facebook import is_facebook_post_url

import minet.facebook as facebook
from minet.cli.utils import with_enricher_and_loading_bar
from minet.cli.crowdtangle.utils import with_crowdtangle_utilities
from minet.crowdtangle.constants import CROWDTANGLE_POST_CSV_HEADERS
from minet.crowdtangle.exceptions import (
    CrowdTanglePostNotFound,
)


@with_crowdtangle_utilities
@with_enricher_and_loading_bar(
    headers=CROWDTANGLE_POST_CSV_HEADERS, title="Retrieving posts", unit="posts"
)
def action(cli_args, client, enricher, loading_bar):
    for row, url in enricher.cells(cli_args.column, with_rows=True):
        with loading_bar.step():
            url = url.strip()

            if not url:
                enricher.writerow(row)
                continue

            url = ensure_protocol(url)

            if not is_facebook_post_url(url):
                enricher.writerow(row)
                continue

            post_id = facebook.post_id_from_url(url)

            if post_id is None:
                enricher.writerow(row)
                continue

            try:
                post = client.post(post_id)
                enricher.writerow(row, post.as_csv_row())
            except CrowdTanglePostNotFound as error:
                enricher.writerow(row)
                loading_bar.print(
                    "\n\n"
                    "CrowdTangle does not have data about the post at this url: {url}\n"
                    "The CrowdTangle API's response is: \n{data}\n".format(
                        url=url, data=error.data
                    )
                )



# =============================================================================
# Minet Twitter Tweet Count CLI Action
# =============================================================================
#
# Logic of the `tw tweet-count` action.
#
from twitter import TwitterHTTPError

from minet.cli.utils import with_enricher_and_loading_bar
from minet.cli.twitter.utils import validate_query_boundaries, with_twitter_client

ITEMS_PER_PAGE = 100

COUNT_FIELDS = ["tweet_count"]
GRANULARIZED_COUNT_FIELDS = ["start_time", "end_time", "tweet_count"]


def get_headers(cli_args):
    return (
        GRANULARIZED_COUNT_FIELDS if cli_args.granularity is not None else COUNT_FIELDS
    )


@with_enricher_and_loading_bar(
    headers=get_headers, title="Counting tweets", unit="queries"
)
@with_twitter_client(api_version="2")
def action(cli_args, client, enricher, loading_bar):
    validate_query_boundaries(cli_args)

    for row, query in enricher.cells(cli_args.column, with_rows=True):
        with loading_bar.step():
            kwargs = {"query": query}

            # Because we are greedy, we want stuff from the beginning of Twitter
            if cli_args.academic and not cli_args.start_time:
                kwargs["start_time"] = "2006-03-21T00:00:00Z"

            if cli_args.start_time:
                kwargs["start_time"] = cli_args.start_time
            if cli_args.end_time:
                kwargs["end_time"] = cli_args.end_time
            if cli_args.since_id:
                kwargs["since_id"] = cli_args.since_id
            if cli_args.until_id:
                kwargs["until_id"] = cli_args.until_id

            kwargs["granularity"] = cli_args.granularity or "day"

            route = (
                ["tweets", "counts", "all"]
                if cli_args.academic
                else ["tweets", "counts", "recent"]
            )

            total_count = 0

            while True:
                try:
                    result = client.call(route, **kwargs)
                except TwitterHTTPError as e:
                    if e.e.code == 404:
                        enricher.writerow(row)
                    else:
                        raise e

                    continue

                for count in result["data"]:
                    total_count += count["tweet_count"]

                    if cli_args.granularity is not None:
                        addendum = [count["start"], count["end"], count["tweet_count"]]
                        enricher.writerow(row, addendum)

                if "next_token" in result["meta"]:
                    kwargs["next_token"] = result["meta"]["next_token"]
                else:
                    break

            if cli_args.granularity is None:
                enricher.writerow(row, [total_count])



# =============================================================================
# Minet Twitter Tweets CLI Action
# =============================================================================
#
# Logic of the `tw tweets` action.
#
from twitwi import normalize_tweet, normalize_tweets_payload_v2, format_tweet_as_csv_row
from twitwi.constants import TWEET_FIELDS, TWEET_EXPANSIONS, TWEET_PARAMS
from twitter import TwitterHTTPError
from ebbe import as_chunks

from minet.cli.utils import with_enricher_and_loading_bar
from minet.cli.twitter.utils import with_twitter_client


@with_enricher_and_loading_bar(
    headers=TWEET_FIELDS, title="Retrieving tweets", unit="tweets"
)
@with_twitter_client()
def action(cli_args, client, enricher, loading_bar):
    def call_client(ids):
        if cli_args.v2:
            kwargs = {
                "ids": ids,
                "expansions": ",".join(TWEET_EXPANSIONS),
                "params": TWEET_PARAMS,
            }
            return client.call(["tweets"], **kwargs)
        else:
            kwargs = {"_id": ids, "tweet_mode": "extended"}
            return client.call(["statuses", "lookup"], **kwargs)

    for chunk in as_chunks(100, enricher.cells(cli_args.column, with_rows=True)):
        tweets = ",".join(row[1] for row in chunk)

        with loading_bar.step(count=len(chunk)):
            try:
                result = call_client(tweets)
            except TwitterHTTPError as e:
                if e.e.code == 404:
                    for row, tweet in chunk:
                        enricher.writerow(row)
                else:
                    raise e

                continue

            indexed_result = {}

            if cli_args.v2:
                normalized_tweets = normalize_tweets_payload_v2(
                    result, collection_source="api"
                )
                for normalized_tweet in normalized_tweets:
                    addendum = format_tweet_as_csv_row(normalized_tweet)
                    indexed_result[normalized_tweet["id"]] = addendum

            else:
                for tweet in result:
                    tweet = normalize_tweet(tweet, collection_source="api")
                    addendum = format_tweet_as_csv_row(tweet)
                    indexed_result[tweet["id"]] = addendum

            for row, tweet in chunk:
                addendum = indexed_result.get(tweet)
                enricher.writerow(row, addendum)



# =============================================================================
# Minet Twitter User Search CLI Action
# =============================================================================
#
# Logic of the `tw user-search` action using API v1 to search users using the
# given query.
#
# Note that the API docs are a bit fuzzy on the matter as it seems that there
# is no good way to handle pagination. As it seems, the last page actually
# still return a maximum number of items, even if it means returning users
# that were already seen, again. Hence the not-straigtforward-looking code
# below to handle pagination in spite of the API's shortcomings.
#
import math
from twitwi import normalize_user, format_user_as_csv_row
from twitwi.constants import USER_FIELDS

from minet.cli.utils import with_enricher_and_loading_bar
from minet.cli.twitter.utils import with_twitter_client

ITEMS_PER_PAGE = 20
MAX_ITEM_COUNT = 1020  # NOTE: API docs say it's 1000, but reality begs to differ
RANGE_UPPER_BOUND = math.ceil(MAX_ITEM_COUNT / ITEMS_PER_PAGE) + 1


@with_enricher_and_loading_bar(
    headers=USER_FIELDS,
    title="Searching users",
    unit="queries",
    nested=True,
    sub_unit="users",
)
@with_twitter_client()
def action(cli_args, client, enricher, loading_bar):

    for row, query in enricher.cells(cli_args.column, with_rows=True):
        with loading_bar.step(query):
            kwargs = {"q": query, "count": ITEMS_PER_PAGE, "include_entities": True}

            already_seen_users = set()

            for page in range(1, RANGE_UPPER_BOUND):
                kwargs["page"] = page

                result = client.call(["users", "search"], **kwargs)
                new_user_count = 0

                for user in result:
                    user = normalize_user(user)

                    if user["id"] in already_seen_users:
                        continue

                    loading_bar.nested_advance()

                    new_user_count += 1
                    already_seen_users.add(user["id"])
                    user_row = format_user_as_csv_row(user)
                    enricher.writerow(row, user_row)

                # If we did not see a single new user, we can safely stop
                # If the number of results is strictly less than the expected number of
                # items per page, we can safely stop
                if new_user_count == 0 or len(result) < ITEMS_PER_PAGE:
                    break



# =============================================================================
# Minet Youtube Captions CLI Action
# =============================================================================
#
# Action retrieving the captions of YouTube videos using the API.
#
from minet.cli.utils import with_enricher_and_loading_bar
from minet.youtube import get_video_captions
from minet.youtube.constants import YOUTUBE_CAPTIONS_CSV_HEADERS


@with_enricher_and_loading_bar(
    headers=YOUTUBE_CAPTIONS_CSV_HEADERS, title="Retrieving captions", unit="videos"
)
def action(cli_args, enricher, loading_bar):
    with loading_bar.step():
        for row, video in enricher.cells(cli_args.column, with_rows=True):
            result = get_video_captions(video, langs=cli_args.lang)

            if result is None:
                continue

            track, lines = result

            prefix = [track.lang, "1" if track.generated else ""]

            for line in lines:
                enricher.writerow(row, prefix + list(line))



from minet.constants import COOKIE_BROWSERS
from minet.cli.utils import with_fatal_errors

from minet.facebook.exceptions import FacebookInvalidCookieError


def fatal_errors_hook(_, e):
    if isinstance(e, FacebookInvalidCookieError):
        if e.target in COOKIE_BROWSERS:
            return 'Could not extract relevant cookie from "%s".' % e.target

        return [
            "Relevant cookie not found.",
            "A Facebook authentication cookie is necessary to be able to scrape Facebook comments.",
            "Use the --cookie flag to choose a browser from which to extract the cookie or give your cookie directly.",
        ]


def with_facebook_fatal_errors(fn):
    return with_fatal_errors(fatal_errors_hook)(fn)


def print_translation_warning_if_needed(loading_bar, translated_langs, post):
    if post.translated_text and post.translated_from not in translated_langs:
        translated_langs.add(post.translated_from)
        lines = [
            "[warning]Found text translated from %s![/warning]" % post.translated_from,
            "Since it means original text may not be entirely retrieved you might want",
            'to edit your Facebook language settings to add "%s" to'
            % post.translated_from,
            'the "Languages you don\'t want to be offered translations for" list here:',
            "https://www.facebook.com/settings/?tab=language",
            "",
        ]

        loading_bar.print(lines)



# =============================================================================
# Minet Twitter API Scraper
# =============================================================================
#
# Twitter public API "scraper".
#
import time
import datetime
from ebbe import with_is_first
from urllib.parse import urlencode, quote
from twitwi import normalize_tweet, normalize_user
from ebbe import getpath, pathgetter
from json import JSONDecodeError

from minet.web import (
    create_pool_manager,
    request,
    create_request_retryer,
    retrying_method,
)
from minet.twitter.constants import (
    TWITTER_PUBLIC_API_DEFAULT_TIMEOUT,
    TWITTER_PUBLIC_API_AUTH_HEADER,
)
from minet.twitter.exceptions import (
    TwitterGuestTokenError,
    TwitterPublicAPIBadRequest,
    TwitterPublicAPIRateLimitError,
    TwitterPublicAPIInvalidResponseError,
    TwitterPublicAPIParsingError,
    TwitterPublicAPIQueryTooLongError,
    TwitterPublicAPIOverCapacityError,
    TwitterPublicAPIHiccupError,
)

# =============================================================================
# Constants
# =============================================================================
TWITTER_PUBLIC_SEARCH_ENDPOINT = "https://api.twitter.com/2/search/adaptive.json"
TWITTER_GUEST_ACTIVATE_ENDPOINT = "https://api.twitter.com/1.1/guest/activate.json"
MAXIMUM_QUERY_LENGTH = 500
DEFAULT_COUNT = 100  # NOTE: the actual upper limit seems to be 20, but I keep 100 just in case it changes in the future, who knows...


# =============================================================================
# Helpers
# =============================================================================
def is_query_too_long(query):
    return len(quote(query)) > MAXIMUM_QUERY_LENGTH


def ensure_guest_token(method):
    def wrapped(self, *args, **kwargs):

        if self.guest_token is None:
            self.acquire_guest_token()

        return method(self, *args, **kwargs)

    return wrapped


def create_cookie_expiration():
    return datetime.datetime.utcfromtimestamp(time.time() + 10800).strftime(
        "%a, %d %b %Y %H:%M:%S GMT"
    )


TWEET_SEARCH_DEFAULT_PARAMS = {
    "include_profile_interstitial_type": "1",
    "include_blocking": "1",
    "include_blocked_by": "1",
    "include_followed_by": "1",
    "include_want_retweets": "1",
    "include_mute_edge": "1",
    "include_can_dm": "1",
    "include_can_media_tag": "1",
    "include_ext_has_nft_avatar": "1",
    "include_ext_is_blue_verified": "1",
    "include_ext_verified_type": "1",
    "skip_status": "1",
    "cards_platform": "Web-12",
    "include_cards": "1",
    "include_ext_alt_text": "true",
    "include_ext_limited_action_results": "false",
    "include_quote_count": "true",
    "include_reply_count": "1",
    "tweet_mode": "extended",
    "include_ext_collab_control": "true",
    "include_ext_views": "true",
    "include_entities": "true",
    "include_user_entities": "true",
    "include_ext_media_color": "true",
    "include_ext_media_availability": "true",
    "include_ext_sensitive_media_warning": "true",
    "include_ext_trusted_friends_metadata": "true",
    "send_error_codes": "true",
    "simple_quoted_tweet": "true",
    "tweet_search_mode": "live",
    "query_source": "spelling_expansion_revert_click",
    "pc": "1",
    "spelling_corrections": "1",
    "include_ext_edit_control": "true",
    "ext": "mediaStats,highlightedLabel,hasNftAvatar,voiceInfo,enrichments,superFollowMetadata,unmentionInfo,editControl,collab_control,vibe",
}

USER_SEARCH_DEFAULT_PARAMS = {
    "include_profile_interstitial_type": "1",
    "include_blocking": "1",
    "include_blocked_by": "1",
    "include_followed_by": "1",
    "include_want_retweets": "1",
    "include_mute_edge": "1",
    "include_can_dm": "1",
    "include_can_media_tag": "1",
    "include_ext_has_nft_avatar": "1",
    "skip_status": "1",
    "cards_platform": "Web-12",
    "include_cards": "1",
    "include_ext_alt_text": "true",
    "include_quote_count": "true",
    "include_reply_count": "1",
    "tweet_mode": "extended",
    "include_entities": "true",
    "include_user_entities": "true",
    "include_ext_media_color": "true",
    "include_ext_media_availability": "true",
    "include_ext_sensitive_media_warning": "true",
    "send_error_codes": "true",
    "simple_quoted_tweet": "true",
    "result_filter": "user",
    "query_source": "typed_query",
    "pc": "1",
    "spelling_corrections": "0",
    "ext": "mediaStats,highlightedLabel,hasNftAvatar,voiceInfo,superFollowMetadata",
}


def forge_search_params(query, target="tweets", count=DEFAULT_COUNT, cursor=None):
    if target == "tweets":
        base = TWEET_SEARCH_DEFAULT_PARAMS
    elif target == "users":
        base = USER_SEARCH_DEFAULT_PARAMS
    else:
        raise TypeError('Invalid target. Expecting "tweets" or "users".')

    params = {**base, "q": query, "count": count}

    if cursor is not None:
        params["cursor"] = cursor

    return urlencode(params, quote_via=quote)


CURSOR_FIRST_POSSIBLE_PATH_GETTER = pathgetter(
    [
        "timeline",
        "instructions",
        0,
        "addEntries",
        "entries",
        -1,
        "content",
        "operation",
        "cursor",
        "value",
    ]
)

CURSOR_SECOND_POSSIBLE_PATH_GETTER = pathgetter(
    [
        "timeline",
        "instructions",
        -1,
        "replaceEntry",
        "entry",
        "content",
        "operation",
        "cursor",
        "value",
    ]
)

CURSOR_USER_PATH_GETTER = pathgetter(
    [
        "timeline",
        "instructions",
        -1,
        "addEntries",
        "entries",
        -1,
        "content",
        "operation",
        "cursor",
        "value",
    ]
)


def extract_cursor_from_tweets_payload(payload):
    found_cursor = CURSOR_FIRST_POSSIBLE_PATH_GETTER(payload)

    if found_cursor is None:
        found_cursor = CURSOR_SECOND_POSSIBLE_PATH_GETTER(payload)

    return found_cursor


def extract_cursor_from_users_payload(payload):
    return CURSOR_USER_PATH_GETTER(payload)


def process_single_tweet(tweet_id, tweet_index, user_index):
    tweet = tweet_index[tweet_id]
    tweet["user"] = user_index[tweet["user_id_str"]]

    # Quoted?
    quoted_id = tweet.get("quoted_status_id_str")

    # TODO: sometimes tweet is not present
    if quoted_id and quoted_id in tweet_index:
        tweet["quoted_status"] = process_single_tweet(
            quoted_id, tweet_index, user_index
        )

    # Retweeted?
    retweeted_id = tweet.get("retweeted_status_id_str")

    # TODO: sometimes tweet is not present
    if retweeted_id and retweeted_id in tweet_index:
        tweet["retweeted_status"] = process_single_tweet(
            retweeted_id, tweet_index, user_index
        )

    return tweet


def tweets_payload_iter(payload):
    tweet_index = payload["globalObjects"]["tweets"]
    user_index = payload["globalObjects"]["users"]

    for instruction in payload["timeline"]["instructions"]:
        if "addEntries" in instruction:
            entries = instruction["addEntries"]["entries"]
        elif "replaceEntry" in instruction:
            entries = [instruction["replaceEntry"]["entry"]]
        else:
            continue

        for entry in entries:
            entry_id = entry["entryId"]

            # Filtering tweets
            if not entry_id.startswith("sq-I-t-") and not entry_id.startswith("tweet-"):
                continue

            tweet_meta = getpath(entry, ["content", "item", "content", "tweet"])

            if tweet_meta is None:
                tweet_meta = getpath(
                    entry, ["content", "item", "content", "tombstone", "tweet"]
                )

            # Parsing error?
            if tweet_meta is None:
                raise TwitterPublicAPIParsingError

            # Skipping ads
            if "promotedMetadata" in tweet_meta:
                continue

            tweet = process_single_tweet(tweet_meta["id"], tweet_index, user_index)

            # Additional metadata
            meta = None

            if tweet is not None:

                if "forwardPivot" in tweet_meta:
                    pivot = tweet_meta["forwardPivot"]

                    meta = {
                        "intervention_text": getpath(pivot, ["text", "text"]),
                        "intervention_type": pivot.get("displayType"),
                        "intervention_url": getpath(pivot, ["landingUrl", "url"]),
                    }

                yield tweet, meta


# =============================================================================
# Main class
# =============================================================================
class TwitterAPIScraper(object):
    def __init__(self):
        self.pool_manager = create_pool_manager(
            timeout=TWITTER_PUBLIC_API_DEFAULT_TIMEOUT, spoof_tls_ciphers=True
        )
        self.reset()

        self.retryer = create_request_retryer(
            min=1,
            additional_exceptions=[
                TwitterPublicAPIRateLimitError,
                TwitterPublicAPIInvalidResponseError,
                TwitterPublicAPIHiccupError,  # TODO: I might want to drop this at some point
            ],
        )

    def reset(self):
        self.guest_token = None
        self.cookie = None

    def request(self, url, headers=None, method="GET"):
        return request(
            url,
            pool_manager=self.pool_manager,
            spoof_ua=True,
            method=method,
            headers=headers,
            known_encoding="utf-8",
        )

    def acquire_guest_token(self):
        headers = {
            "Authorization": TWITTER_PUBLIC_API_AUTH_HEADER,
            "Accept-Language": "en-US,en;q=0.5",
        }

        response = self.request(TWITTER_GUEST_ACTIVATE_ENDPOINT, headers, method="POST")

        if response.status >= 400:
            raise TwitterPublicAPIInvalidResponseError

        try:
            api_token_response = response.json()
        except JSONDecodeError:
            raise TwitterPublicAPIInvalidResponseError

        guest_token = api_token_response.get("guest_token")
        if guest_token is None:
            raise TwitterGuestTokenError

        self.guest_token = guest_token
        self.cookie = response.headers[
            "Set-Cookie"
        ] + ", gt=%s; Domain=.twitter.com; Path=/; Secure ; Expires=%s" % (
            guest_token,
            create_cookie_expiration(),
        )

    @retrying_method()
    @ensure_guest_token
    def request_search(self, url):
        headers = {
            "Authorization": TWITTER_PUBLIC_API_AUTH_HEADER,
            "X-Guest-Token": self.guest_token,
            "Cookie": self.cookie,
            "Accept-Language": "en-US,en;q=0.5",
        }

        response = self.request(url, headers=headers)

        if response.status in [403, 429]:
            self.reset()
            raise TwitterPublicAPIRateLimitError

        data = response.json()

        if response.status >= 400:
            error = getpath(data, ["errors", 0])

            if error is not None and response.status == 400 and error.get("code") == 47:
                raise TwitterPublicAPIBadRequest

            if error is not None and error.get("code") == 130:
                raise TwitterPublicAPIOverCapacityError

            raise TwitterPublicAPIInvalidResponseError

        return data

    def request_tweet_search(self, query, cursor=None, refs=None, dump=False):
        params = forge_search_params(query, cursor=cursor, target="tweets")
        url = "%s?%s" % (TWITTER_PUBLIC_SEARCH_ENDPOINT, params)

        data = self.request_search(url)

        next_cursor = extract_cursor_from_tweets_payload(data)
        tweets = []

        if dump:
            return data

        for tweet, meta in tweets_payload_iter(data):
            result = normalize_tweet(
                tweet,
                extract_referenced_tweets=refs is not None,
                collection_source="scraping",
            )

            if refs is not None:
                for is_first, extracted_tweet in with_is_first(result):

                    # Casting to int64 to save up memory
                    id_int64 = int(extracted_tweet["id"])

                    if id_int64 in refs:
                        continue

                    if is_first:
                        tweets.append((extracted_tweet, meta))
                    else:
                        tweets.append((extracted_tweet, None))

                    refs.add(id_int64)
            else:
                tweets.append((result, meta))

        # Attempting to fix Twitter's public-facing API recent hiccups (#316):
        # It seems that sometimes the API returns an empty response, containing
        # the same cursor as before, in which case we should retry...
        # NOTE: in fact, the real issue lies elsewhere as the way to hit the API
        # changed slightly and was causing our issues. But we cannot rely
        # on this condition now because it will degenerate to a retry loop
        # when hitting the last available tweets of the query.
        # NOTE: since those API changes, we cannot get 100 results at once anymore
        # as the upper limit seems to be 20 now :'(

        # if not tweets and cursor == next_cursor:
        #     raise TwitterPublicAPIHiccupError

        return next_cursor, tweets

    def request_user_search(self, query, cursor=None, dump=False):
        params = forge_search_params(query, cursor=cursor, target="users")
        url = "%s?%s" % (TWITTER_PUBLIC_SEARCH_ENDPOINT, params)

        data = self.request_search(url)

        next_cursor = extract_cursor_from_users_payload(data)

        if dump:
            return data

        users = [
            normalize_user(user) for user in data["globalObjects"]["users"].values()
        ]

        return next_cursor, users

    def search_tweets(
        self, query, limit=None, include_referenced_tweets=False, with_meta=False
    ):

        if is_query_too_long(query):
            raise TwitterPublicAPIQueryTooLongError

        cursor = None
        i = 0

        refs = set() if include_referenced_tweets else None

        while True:
            new_cursor, tweets = self.request_tweet_search(query, cursor, refs=refs)

            for tweet, meta in tweets:
                if with_meta:
                    yield tweet, meta

                else:
                    yield tweet

                i += 1

                if limit is not None and i >= limit:
                    return

            # NOTE: we cannot stop when no tweets are found anymore because
            # of Twitter's public facing API's strange hiccups.
            # Comparing cursors seems to be the only good option now...
            if new_cursor is None or new_cursor == cursor:
                return

            cursor = new_cursor

    def search_users(self, query, limit=None):
        if is_query_too_long(query):
            raise TwitterPublicAPIQueryTooLongError

        cursor = None
        i = 0

        while True:
            new_cursor, users = self.request_user_search(query, cursor)

            if not users:
                return

            for user in users:
                yield user

                i += 1

                if limit is not None and i >= limit:
                    return

            if new_cursor is None or new_cursor == cursor:
                return

            cursor = new_cursor



from minet.cli.utils import with_fatal_errors
from minet.hyphe.exceptions import HypheCorpusAuthenticationError

FATAL_ERRORS = {
    HypheCorpusAuthenticationError: [
        "Wrong password for corpus!",
        "Don't forget to provide a password for this corpus using --password",
    ]
}


def with_hyphe_fatal_errors(fn):
    return with_fatal_errors(FATAL_ERRORS)(fn)



# =============================================================================
# Minet Twitter CLI Action
# =============================================================================
#
# Logic of the `tw` action.
#
from casanova import RowCountResumer, BatchResumer

from minet.cli.argparse import (
    command,
    subcommand,
    ConfigAction,
    TimezoneType,
    TimestampType,
)
from minet.cli.exceptions import InvalidArgumentsError

TWITTER_API_COMMON_ARGUMENTS = [
    {
        "flag": "--api-key",
        "help": "Twitter API key.",
        "rc_key": ["twitter", "api_key"],
        "action": ConfigAction,
    },
    {
        "flag": "--api-secret-key",
        "help": "Twitter API secret key.",
        "rc_key": ["twitter", "api_secret_key"],
        "action": ConfigAction,
    },
    {
        "flag": "--access-token",
        "help": "Twitter API access token.",
        "rc_key": ["twitter", "access_token"],
        "action": ConfigAction,
    },
    {
        "flag": "--access-token-secret",
        "help": "Twitter API access token secret.",
        "rc_key": ["twitter", "access_token_secret"],
        "action": ConfigAction,
    },
]

V2_ARGUMENT = {
    "flag": "--v2",
    "help": "Whether to use latest Twitter API v2 rather than v1.1.",
    "action": "store_true",
}

IDS_ARGUMENT = {
    "flag": "--ids",
    "help": "Whether your users are given as ids rather than screen names.",
    "action": "store_true",
}

ACADEMIC_ARGUMENT = {
    "flag": "--academic",
    "help": "Flag to add if you want to use your academic research access (in order to search the complete history of public tweets).",
    "action": "store_true",
}

COMMON_V2_SEARCH_ARGUMENTS = [
    {
        "flag": "--since-id",
        "help": "Will return tweets with ids that are greater than the specified id. Takes precedence over --start-time.",
        "type": int,
    },
    {
        "flag": "--until-id",
        "help": "Will return tweets that are older than the tweet with the specified id.",
        "type": int,
    },
    {
        "flag": "--start-time",
        "help": "The oldest UTC stamp from which the tweets will be provided. The date should have the format : YYYY-MM-DDTHH:mm:ssZ",
    },
    {
        "flag": "--end-time",
        "help": "The UTC stamp to which the tweets will be provided. The date should have the format : YYYY-MM-DDTHH:mm:ssZ",
    },
]


def check_credentials(cli_args):

    # Credentials are required to be able to access the API
    if (
        not cli_args.api_key
        or not cli_args.api_secret_key
        or not cli_args.access_token
        or not cli_args.access_token_secret
    ):
        raise InvalidArgumentsError(
            [
                "Full credentials are required to access Twitter API.",
                "You can provide them using various CLI arguments:",
                "    --api-key",
                "    --api-secret-key",
                "    --access-token",
                "    --access-token-secret",
            ]
        )


def twitter_api_subcommand(*args, arguments=[], **kwargs):
    return subcommand(
        *args,
        arguments=arguments + TWITTER_API_COMMON_ARGUMENTS,
        resolve=check_credentials,
        **kwargs
    )


TWITTER_ATTRITION_SUBCOMMAND = twitter_api_subcommand(
    "attrition",
    "minet.cli.twitter.attrition",
    title="Minet Twitter Attrition Command",
    description="""
        Using Twitter API to find whether batches of tweets are still
        available today and if they aren't, attempt to find a reason why.

        This command relies on tweet ids or tweet urls. We recommand to add `--user` and
        the tweet's user id to the command if you can, as more information can
        be obtained when the user id (or the full url) is known.

        The same can be said about retweet information and the `--retweeted-id` flag.

        The command will output a report similar to the input file and
        containing an additional column named "current_tweet_status" that can take
        the following values:

            - "available_tweet": tweet is still available.
            - "user_or_tweet_deleted": tweet was deleted or its author was deactivated. To know whether it is one or the other reason
                                        for unavailability that is the right one, add --user to the command.
            - "suspended_user": tweet cannot be found because its user is suspended.
            - "deactivated_user": tweet cannot be found because its user is deactivated.
            - "deactivated_or_renamed_user": tweet cannot be found because its user is either deactivated or changed its screen name
                                                (only when using tweet urls or tweet ids and screen names instead of user ids).
            - "protected_user": tweet cannot be found because its user is protected.
            - "censored_tweet": tweet is unavailable because it was consored by Twitter.
            - "blocked_by_tweet_author": tweet cannot be found because you were blocked by its author.
            - "unavailable_tweet": tweet is not available, which means it was
                                    deleted by its user.
            - "unavailable_retweet": retweet is not available, meaning that the user
                                        cancelled their retweet.
            - "unavailable_retweeted_tweet": the retweeted tweet is unavailable,
                                                meaning it was either deleted by its original
                                                user or the original user was deactivated.
            - "censored_retweeted_tweet": the original tweet was censored by Twitter, making the retweet unavailable.
            - "protected_retweeted_user": tweet cannot be found because it is a retweet of a protected user.
            - "suspended_retweeted_user": tweet cannot be found because it is a retweet of a suspended user.
            - "blocked_by_original_tweet_author": tweet cannot be found because it is a retweet of a user who blocked you.
    """,
    epilog="""
        Examples:

        . Finding out if tweets in a CSV files are still available or not using tweet ids:
            $ minet tw attrition tweet_url -i deleted_tweets.csv > attrition-report.csv

        . Finding out if tweets are still available or not using tweet & user ids:
            $ minet tw attrition tweet_id -i deleted_tweets.csv --user user_id --ids > attrition-report.csv
    """,
    variadic_input={
        "dummy_column": "tweet_url_or_id",
        "item_label": "tweet url or id",
        "item_label_plural": "tweet urls or ids",
    },
    resumer=RowCountResumer,
    arguments=[
        {
            "flag": "--user",
            "help": "Name of the column containing the tweet's author (given as ids or screen names). This is useful to have more information on a tweet's unavailability.",
        },
        {
            "flag": "--retweeted-id",
            "help": "Name of the column containing the ids of the original tweets in case the tweets no longer available were retweets.",
        },
        IDS_ARGUMENT,
    ],
)

TWITTER_FOLLOWERS_SUBCOMMAND = twitter_api_subcommand(
    "followers",
    "minet.cli.twitter.followers",
    title="Minet Twitter Followers Command",
    description="""
        Retrieve followers, i.e. followed users, of given user.
    """,
    epilog="""
        Examples:

        . Getting followers of a list of user:
            $ minet tw followers screen_name -i users.csv > followers.csv
    """,
    variadic_input={
        "dummy_column": "user",
        "item_label": "Twitter account screen name or id",
        "item_label_plural": "Twitter account screen names or ids",
    },
    resumer=BatchResumer,
    resumer_kwargs=lambda args: ({"value_column": args.column}),
    arguments=[
        IDS_ARGUMENT,
        V2_ARGUMENT,
    ],
)

TWITTER_FRIENDS_SUBCOMMAND = twitter_api_subcommand(
    "friends",
    "minet.cli.twitter.friends",
    title="Minet Twitter Friends Command",
    description="""
        Retrieve friends, i.e. followed users, of given user.
    """,
    epilog="""
        Examples:

        . Getting friends of a list of user:
            $ minet tw friends screen_name -i users.csv > friends.csv
    """,
    variadic_input={
        "dummy_column": "user",
        "item_label": "Twitter account screen name or id",
        "item_label_plural": "Twitter account screen names or ids",
    },
    resumer=BatchResumer,
    resumer_kwargs=lambda args: ({"value_column": args.column}),
    arguments=[
        IDS_ARGUMENT,
        V2_ARGUMENT,
    ],
)

TWITTER_LIST_FOLLOWERS_SUBCOMMAND = twitter_api_subcommand(
    "list-followers",
    "minet.cli.twitter.list_followers",
    title="Minet Twitter List Followers Command",
    description="""
        Retrieve followers of given list using Twitter API v2.
    """,
    epilog="""
        Examples:

        . Getting followers of a list of lists:
            $ minet tw list-followers id -i lists.csv > followers.csv
    """,
    variadic_input={
        "dummy_column": "list",
        "item_label": "Twitter list id or url",
        "item_label_plural": "Twitter list ids or urls",
    },
)

TWITTER_LIST_MEMBERS_SUBCOMMAND = twitter_api_subcommand(
    "list-members",
    "minet.cli.twitter.list_members",
    title="Minet Twitter List Members Command",
    description="""
        Retrieve members of given list using Twitter API v2.
    """,
    epilog="""
        Examples:

        . Getting members of a list of lists:
            $ minet tw list-members id -i lists.csv > members.csv
    """,
    variadic_input={
        "dummy_column": "list",
        "item_label": "Twitter list id or url",
        "item_label_plural": "Twitter list ids or urls",
    },
)

TWITTER_RETWEETERS_SUBCOMMAND = twitter_api_subcommand(
    "retweeters",
    "minet.cli.twitter.retweeters",
    title="Minet Twitter Retweeters Command",
    description="""
        Retrieve retweeters of given tweet using Twitter API v2.
    """,
    epilog="""
        Examples:

        . Getting the users who retweeted a list of tweets:
            $ minet tw retweeters tweet_id -i tweets.csv > retweeters.csv
    """,
    variadic_input={"dummy_column": "tweet_id", "item_label": "tweet id"},
)

TWITTER_SCRAPE_SUBCOMMAND = subcommand(
    "scrape",
    "minet.cli.twitter.scrape",
    title="Minet Twitter Scrape Command",
    description="""
        Scrape Twitter's public facing search API to collect tweets or users.

        Be sure to check Twitter's advanced search to check what kind of
        operators you can use to tune your queries (time range, hashtags,
        mentions, boolean etc.):
        https://twitter.com/search-advanced?f=live

        Useful operators include "since" and "until" to search specific
        time ranges like so: "since:2014-01-01 until:2017-12-31", or "from" that
        let you search all tweets from a given user like so: "from:medialab_ScPo".

        BEWARE: the web search results seem to become inconsistent when
        queries return vast amounts of tweets. In which case you are
        strongly advised to segment your queries using temporal filters.
    """,
    epilog="""
        Examples:

        . Collecting the latest 500 tweets of a given Twitter user:
            $ minet tw scrape tweets "from:@jack" --limit 500 > tweets.csv

        . Collecting the tweets from multiple Twitter queries listed in a CSV file:
            $ minet tw scrape tweets query -i queries.csv > tweets.csv

        . Templating the given CSV column to query tweets by users:
            $ minet tw scrape tweets user -i users.csv --query-template 'from:@{value}' > tweets.csv

        . Collecting users with "adam" in their user_name or user_description:
            $ minet tw scrape users adam > users.csv

        Tips:

        - You can add a "OR @aNotExistingHandle" to your query to avoid searching
          for your query terms in usernames or handles. Note that this is a temporary hack
          which might stop working at any time so be sure to double check before relying on
          this trick.

            $ minet tw scrape tweets "keyword OR @anObviouslyNotExistingHandle"

          For more information see the related discussion here:
          https://webapps.stackexchange.com/questions/127425/how-to-exclude-usernames-and-handles-while-searching-twitter
    """,
    variadic_input={"dummy_column": "query", "item_label_plural": "queries"},
    arguments=[
        {
            "name": "items",
            "help": "What to scrape. Currently only allows for `tweets` or `users`.",
            "choices": ["tweets", "users"],
        },
        {
            "flag": "--include-refs",
            "help": "Whether to emit referenced tweets (quoted, retweeted & replied) in the CSV output. Note that it consumes a memory proportional to the total number of unique tweets retrieved.",
            "action": "store_true",
        },
        {
            "flags": ["-l", "--limit"],
            "help": "Maximum number of tweets or users to collect per query.",
            "type": int,
        },
        {
            "flag": "--query-template",
            "help": "Query template. Can be useful for instance to change a column of twitter user screen names into from:@user queries.",
        },
    ],
)

TWITTER_TWEET_COUNT_SUBCOMMAND = twitter_api_subcommand(
    "tweet-count",
    "minet.cli.twitter.tweet_count",
    title="Minet Twitter Tweets Count Command",
    description="""
        Count the number of tweets matching the given query using Twitter's
        latest API v2. The count's granularity can be at the level of tweets
        per day, per hour, or per minute.

        This will only return result for the last 8 days only, unless
        you have Academic Research access in which case you
        can use the --academic flag to retrieve the full count.

        Be advised that, for now, results are returned unordered
        by Twitter's API if you choose a time granularity for the
        results.
    """,
    epilog="""
        Examples:

        . Counting tweets using "cancer" as a query:
            $ minet tw tweet-count cancer

        . Running multiple queries in series:
            $ minet tw tweet-count query -i queries.csv > counts.csv

        . Number of tweets matching the query per day:
            $ minet tw tweet-count "query" --granularity day > counts.csv
    """,
    variadic_input={"dummy_column": "query", "item_label_plural": "queries"},
    arguments=[
        {
            "flag": "--granularity",
            "help": "Granularity used to group the data by. Defaults to day.",
            "choices": ["minute", "hour", "day"],
        },
        *COMMON_V2_SEARCH_ARGUMENTS,
        ACADEMIC_ARGUMENT,
    ],
)

TWITTER_TWEET_DATE_SUBCOMMAND = subcommand(
    "tweet-date",
    "minet.cli.twitter.tweet_date",
    title="Minet Twitter Tweet Date Command",
    description="""
        Getting timestamp and date from tweet url or id.
    """,
    epilog="""
        Examples:

            $ minet tw tweet-date url -i tweets.csv --timezone 'Europe/Paris'> tweets_timestamp_date.csv
    """,
    variadic_input={
        "dummy_column": "tweet",
        "item_label": "tweet url or id",
        "item_label_plural": "tweet urls or ids",
    },
    arguments=[
        {
            "flag": "--timezone",
            "help": "Timezone for the date, for example 'Europe/Paris'. Default to UTC.",
            "type": TimezoneType(),
        }
    ],
)

TWITTER_TWEET_SEARCH_SUBCOMMAND = twitter_api_subcommand(
    "tweet-search",
    "minet.cli.twitter.tweet_search",
    title="Minet Twitter Tweets Search Command",
    description="""
        Search Twitter tweets using API v2.

        This will only return the last 8 days of results maximum per query (unless you have Academic Research access).

        To search the full archive of public tweets, use --academic if you have academic research access.
    """,
    epilog="""
        Examples:

        . Searching tweets using "cancer" as a query:
            $ minet tw tweet-search cancer > tweets.csv

        . Running multiple queries in series:
            $ minet tw tweet-search query -i queries.csv > tweets.csv
    """,
    variadic_input={"dummy_column": "query", "item_label_plural": "queries"},
    arguments=[
        *COMMON_V2_SEARCH_ARGUMENTS,
        ACADEMIC_ARGUMENT,
        {
            "flag": "--sort-order",
            "help": 'How to sort retrieved tweets. Defaults to "recency".',
            "choices": ["recency", "relevancy"],
            "default": "recency",
        },
    ],
)

TWITTER_TWEETS_SUBCOMMAND = twitter_api_subcommand(
    "tweets",
    "minet.cli.twitter.tweets",
    title="Minet Twitter Tweets Command",
    description="""
        Collecting tweet metadata from the given tweet ids, using the API.
    """,
    epilog="""
        Examples:

        . Getting metadata from tweets in a CSV file:
            $ minet tw tweets tweet_id -i tweets.csv > tweets_metadata.csv
    """,
    variadic_input={"dummy_column": "tweet_id", "item_label": "tweet id"},
    resumer=RowCountResumer,
    arguments=[V2_ARGUMENT],
)

TWITTER_USER_SEARCH_SUBCOMMAND = twitter_api_subcommand(
    "user-search",
    "minet.cli.twitter.user_search",
    title="Minet Twitter Users Search Command",
    description="""
        Search Twitter users using API v1.

        This will only return ~1000 results maximum per query
        so you might want to find a way to segment your inquiry
        into smaller queries to find more users.
    """,
    epilog="""
        Examples:

        . Searching user using "cancer" as a query:
            $ minet tw user-search cancer > users.csv

        . Running multiple queries in series:
            $ minet tw user-search query -i queries.csv > users.csv
    """,
    variadic_input={"dummy_column": "query", "item_label_plural": "queries"},
)

TWITTER_USER_TWEETS_SUBCOMMAND = twitter_api_subcommand(
    "user-tweets",
    "minet.cli.twitter.user_tweets",
    title="Minet Twitter User Tweets Command",
    description="""
        Retrieve the last ~3200 tweets, including retweets from
        the given Twitter users, using the API.
    """,
    epilog="""
        Examples:

        . Getting tweets from users in a CSV file:
            $ minet tw user-tweets screen_name -i users.csv > tweets.csv
    """,
    variadic_input={
        "dummy_column": "user",
        "item_label": "Twitter account screen name or id",
        "item_label_plural": "Twitter account screen names or ids",
    },
    arguments=[
        IDS_ARGUMENT,
        {
            "flag": "--min-date",
            "help": "Whether to add a date to stop at for user's tweets retrieval. UTC date should have the following format : YYYY-MM-DD",
            "type": TimestampType(),
        },
        {
            "flag": "--exclude-retweets",
            "help": "Whether to exclude retweets from the output.",
            "action": "store_true",
        },
        V2_ARGUMENT,
    ],
)

TWITTER_USERS_SUBCOMMAND = twitter_api_subcommand(
    "users",
    "minet.cli.twitter.users",
    title="Minet Twitter Users Command",
    description="""
        Retrieve Twitter user metadata using the API.
    """,
    epilog="""
        Examples:

        . Getting friends of a list of user:
            $ minet tw users screen_name -i users.csv > data_users.csv
    """,
    resumer=RowCountResumer,
    variadic_input={"dummy_column": "user", "item_label": "Twitter user"},
    arguments=[
        IDS_ARGUMENT,
        V2_ARGUMENT,
    ],
)

TWITTER_COMMAND = command(
    "twitter",
    "minet.cli.twitter",
    "Minet Twitter Command",
    aliases=["tw"],
    description="""
        Gather data from Twitter.
    """,
    subcommands=[
        TWITTER_ATTRITION_SUBCOMMAND,
        TWITTER_FOLLOWERS_SUBCOMMAND,
        TWITTER_FRIENDS_SUBCOMMAND,
        TWITTER_LIST_FOLLOWERS_SUBCOMMAND,
        TWITTER_LIST_MEMBERS_SUBCOMMAND,
        TWITTER_RETWEETERS_SUBCOMMAND,
        TWITTER_SCRAPE_SUBCOMMAND,
        TWITTER_TWEET_COUNT_SUBCOMMAND,
        TWITTER_TWEET_DATE_SUBCOMMAND,
        TWITTER_TWEET_SEARCH_SUBCOMMAND,
        TWITTER_TWEETS_SUBCOMMAND,
        TWITTER_USER_SEARCH_SUBCOMMAND,
        TWITTER_USER_TWEETS_SUBCOMMAND,
        TWITTER_USERS_SUBCOMMAND,
    ],
)
